@@ -7,9 +7,12 @@ By default, the entire registry is rebuilt. If you want to do a rebuild of
 changed packages only, you can pass --previous-package-list(-json) with the old
 package list.
 """
+from __future__ import annotations
+
 import argparse
 import collections
 import contextlib
+import inspect
 import itertools
 import json
 import math
@@ -19,18 +22,12 @@ import sys
 import tempfile
 from datetime import datetime
 from typing import Any
-from typing import Dict
 from typing import Generator
 from typing import IO
 from typing import Iterator
-from typing import List
 from typing import NamedTuple
-from typing import Optional
 from typing import Sequence
-from typing import Set
-from typing import Tuple
 
-import distlib.wheel
 import jinja2
 import packaging.utils
 import packaging.version
@@ -48,13 +45,10 @@ def remove_extension(name: str) -> str:
 
 def guess_name_version_from_filename(
         filename: str,
-) -> Tuple[str, Optional[str]]:
+) -> tuple[str, str | None]:
     if filename.endswith('.whl'):
-        m = distlib.wheel.FILENAME_RE.match(filename)
-        if m is not None:
-            return m.group('nm'), m.group('vn')
-        else:
-            raise ValueError(f'Invalid package name: {filename}')
+        name, version, _, _ = packaging.utils.parse_wheel_filename(filename)
+        return name, str(version)
     else:
         # These don't have a well-defined format like wheels do, so they are
         # sort of "best effort", with lots of tests to back them up.
@@ -85,20 +79,20 @@ def guess_name_version_from_filename(
 class Package(NamedTuple):
     filename: str
     name: str
-    version: Optional[str]
+    version: str | None
     parsed_version: packaging.version.Version
-    hash: Optional[str]
-    requires_dist: Optional[Tuple[str, ...]]
-    requires_python: Optional[str]
-    upload_timestamp: Optional[int]
-    uploaded_by: Optional[str]
+    hash: str | None
+    requires_dist: tuple[str, ...] | None
+    requires_python: str | None
+    upload_timestamp: int | None
+    uploaded_by: str | None
 
-    def __lt__(self, other: Tuple[Any, ...]) -> bool:
+    def __lt__(self, other: tuple[Any, ...]) -> bool:
         assert isinstance(other, Package), type(other)
         return self.sort_key < other.sort_key
 
     @property
-    def sort_key(self) -> Tuple[str, packaging.version.Version, str]:
+    def sort_key(self) -> tuple[str, packaging.version.Version, str]:
         """Sort key for a filename."""
         return (
             self.name,
@@ -147,8 +141,8 @@ class Package(NamedTuple):
         else:
             return 'sdist'
 
-    def json_info(self, base_url: str) -> Dict[str, Any]:
-        ret: Dict[str, Any] = {
+    def json_info(self, base_url: str) -> dict[str, Any]:
+        ret: dict[str, Any] = {
             'filename': self.filename,
             'url': self.url(base_url, include_hash=False),
             'requires_python': self.requires_python,
@@ -161,17 +155,25 @@ class Package(NamedTuple):
             ret['digests'] = {algo: h}
         return ret
 
+    def input_json(self) -> dict[str, Any]:
+        """A dict suitable for json lines."""
+        return {
+            k: getattr(self, k)
+            for k in inspect.getfullargspec(self.create).kwonlyargs
+            if getattr(self, k) is not None
+        }
+
     @classmethod
     def create(
             cls,
             *,
             filename: str,
-            hash: Optional[str] = None,
-            requires_dist: Optional[Sequence[str]] = None,
-            requires_python: Optional[str] = None,
-            upload_timestamp: Optional[int] = None,
-            uploaded_by: Optional[str] = None,
-    ) -> 'Package':
+            hash: str | None = None,
+            requires_dist: Sequence[str] | None = None,
+            requires_python: str | None = None,
+            upload_timestamp: int | None = None,
+            uploaded_by: str | None = None,
+    ) -> Package:
         if not re.match(r'[a-zA-Z0-9_\-\.\+]+$', filename) or '..' in filename:
             raise ValueError(f'Unsafe package name: {filename}')
 
@@ -217,10 +219,10 @@ IMPORTANT_METADATA_FOR_INFO = frozenset((
 ))
 
 
-def _package_json(sorted_files: List[Package], base_url: str) -> Dict[str, Any]:
+def _package_json(sorted_files: list[Package], base_url: str) -> dict[str, Any]:
     # https://warehouse.pypa.io/api-reference/json.html
     # note: the full api contains much more, we only output the info we have
-    by_version: Dict[str, List[Package]] = collections.defaultdict(list)
+    by_version: dict[str, list[Package]] = collections.defaultdict(list)
     for file in sorted_files:
         if file.version is not None:
             by_version[file.version].append(file)
@@ -266,8 +268,8 @@ class Settings(NamedTuple):
 
 
 def build_repo(
-        packages: Dict[str, Set[Package]],
-        previous_packages: Optional[Dict[str, Set[Package]]],
+        packages: dict[str, set[Package]],
+        previous_packages: dict[str, set[Package]] | None,
         settings: Settings,
 ) -> None:
     simple = os.path.join(settings.output_dir, 'simple')
@@ -380,16 +382,22 @@ def build_repo(
             ),
         ))
 
+    # /packages.json
+    # Always rebuild (we would have short circuited already if nothing changed).
+    with atomic_write(os.path.join(settings.output_dir, 'packages.json')) as f:
+        for package in itertools.chain.from_iterable(packages.values()):
+            f.write(f'{json.dumps(package.input_json())}\n')
 
-def _lines_from_path(path: str) -> List[str]:
+
+def _lines_from_path(path: str) -> list[str]:
     f = sys.stdin if path == '-' else open(path)
     return f.read().splitlines()
 
 
 def _create_packages(
-        package_infos: Iterator[Dict[str, Any]],
-) -> Dict[str, Set[Package]]:
-    packages: Dict[str, Set[Package]] = collections.defaultdict(set)
+        package_infos: Iterator[dict[str, Any]],
+) -> dict[str, set[Package]]:
+    packages: dict[str, set[Package]] = collections.defaultdict(set)
     for package_info in package_infos:
         try:
             package = Package.create(**package_info)
@@ -402,15 +410,15 @@ def _create_packages(
     return packages
 
 
-def package_list(path: str) -> Dict[str, Set[Package]]:
+def package_list(path: str) -> dict[str, set[Package]]:
     return _create_packages({'filename': line} for line in _lines_from_path(path))
 
 
-def package_list_json(path: str) -> Dict[str, Set[Package]]:
+def package_list_json(path: str) -> dict[str, set[Package]]:
     return _create_packages(json.loads(line) for line in _lines_from_path(path))
 
 
-def main(argv: Optional[Sequence[str]] = None) -> int:
+def main(argv: Sequence[str] | None = None) -> int:
     parser = argparse.ArgumentParser(
         description=__doc__,
         formatter_class=argparse.RawTextHelpFormatter,
